# https://github.com/jpanetta/MeshFEM/blob/master/python/init_template.py
################################################################################
#  auto-generated from /Users/yijiangh/conmech/src/pyconmech/init_template.py
################################################################################
import sys as _sys
_sys.path.insert(0, '/Users/yijiangh/conmech/src/pyconmech')

from _pystiffness_checker import _StiffnessChecker
