"""
********************************************************************************
pyconmech.database
********************************************************************************

.. currentmodule:: pyconmech.database

This package contains some commonly used material properties, cross section properties
for conmech.

"""

from __future__ import absolute_import

from .material_properties import *